import logging
import os
from typing import Any, ClassVar, Dict, Optional

from requests import HTTPError, Response, Session, codes
from requests.exceptions import SSLError

from .exceptions import DhlmexException
from .resources import Guide, PostCode, Resource

PREPAID_URL = 'https://prepaid.dhl.com.mx/Prepago'
USER_AGENT = (
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/537.36 '
    '(KHTML, like Gecko) Chrome/75.0.3770.142 Safari/537.36'
)
DHL_CERT = 'prepaid-dhl-com-mx.pem'

logging.basicConfig(level=logging.DEBUG)


class Client:

    base_url: ClassVar[str] = PREPAID_URL
    headers: Dict[str, str]
    session: Session

    # resources
    guides: ClassVar = Guide
    post_codes: ClassVar = PostCode

    def __init__(
        self, username: Optional[str] = None, password: Optional[str] = None,
    ):

        username = username or os.environ['DHLMEX_USERNAME']
        password = password or os.environ['DHLMEX_PASSWORD']
        self.session = Session()
        self.session.headers['User-Agent'] = USER_AGENT
        if os.getenv('DEBUG'):
            logging.debug(f'Client using Charles certificate')
            self.session.verify = DHL_CERT
        self._login(username, password)

        Resource._client = self

    def _login(self, username: str, password: str) -> Response:
        try:
            self.get('/')  # Initialize cookies
            endpoint = '/jsp/app/login/login.xhtml'
            data = {
                'AJAXREQUEST': '_viewRoot',
                'j_id6': 'j_id6',
                'j_id6:j_id20': username,
                'j_id6:j_id22': password,
                'javax.faces.ViewState': 'j_id1',
                'j_id6:j_id29': 'j_id6:j_id29',
            }
            resp = self.post(endpoint, data)
        except HTTPError as httpe:
            if 'Su sesión ha caducado' in resp.text:
                raise DhlmexException('Session has expired')
            else:
                raise httpe
        except SSLError:
            raise DhlmexException('Client on debug, but Charles not running')
        # DHL always return 200 although there is an existing session
        if 'Ya existe una sesión' in resp.text:
            raise DhlmexException(
                f'There is an exisiting session on DHL for {username}'
            )
        if 'Verifique su usuario' in resp.text:
            raise DhlmexException('Invalid credentials')
        return resp

    def _logout(self) -> Response:
        endpoint = '/jsp/app/inicio/inicio.xhtml'
        resp = self.post(endpoint, {})
        if 'Login / Admin' in resp.text:
            return resp  # No need to logout
        data = Resource.get_data(
            resp, Resource._actions['close'],
        )  # Obtain headers to end properly the session
        try:
            resp = self.post(endpoint, data)
        except HTTPError as httpe:
            if 'Su sesión ha caducado' in httpe.response.text:
                resp = Response()
                resp.status_code = codes.ok
                return resp
            else:
                raise httpe
        return resp

    def get(self, endpoint: str, **kwargs: Any) -> Response:
        return self.request('get', endpoint, {}, **kwargs)

    def post(
        self, endpoint: str, data: Dict[str, str], **kwargs: Any
    ) -> Response:
        return self.request('post', endpoint, data, **kwargs)

    def request(
        self, method: str, endpoint: str, data: Dict[str, str], **kwargs: Any,
    ) -> Response:
        url = self.base_url + endpoint
        response = self.session.request(method, url, data=data, **kwargs)
        self._check_response(response)
        return response

    @staticmethod
    def _check_response(response: Response) -> None:
        if response.ok:
            return
        response.raise_for_status()



from typing import Dict

from dhlmex.exceptions import DhlmexException
from dhlmex.resources.destination import Destination
from dhlmex.resources.origin import Origin

from .base import Resource


class PostCode(Resource):
    @classmethod
    def validate_postal_codes(
        cls, origin: Origin, destination: Destination, view_state: str
    ):
        post_code = cls()
        return post_code._validate_postal_codes(
            origin, destination, view_state
        )

    def _validate_postal_codes(
        self, origin: Origin, destination: Destination, view_state: str
    ) -> Dict:
        fill_data = {
            'AJAXREQUEST': '_viewRoot',
            'datos': 'datos',
            'datos:j_id10': 'j_id11',
            'datos:j_id15': '',
            'datos:j_id19': '',
            'datos:emailOrigen': '',
            'datos:j_id24': '',
            'datos:j_id28': '',
            'datos:j_id30': '',
            'datos:j_id32': '',
            'datos:j_id36': origin.postal_code,
            'datos:j_id41': '',
            'datos:j_id45': '',
            'datos:j_id49': '',
            'datos:j_id54': '',
            'datos:j_id58': '',
            'datos:emailDestino': '',
            'datos:j_id63': '',
            'datos:j_id67': '',
            'datos:j_id69': '',
            'datos:j_id71': '',
            'datos:j_id75': '',
            'datos:j_id80': '',
            'datos:j_id84': '',
            'datos:j_id88': '',
            'datos:j_id93': '',
            'datos:j_id95': '',
            'javax.faces.ViewState': view_state,
            'datos:j_id37': 'datos:j_id37',
        }
        resp = self._client.post(self._urls['capture'], fill_data)
        if 'Código Postal válido' in resp.text:
            fill_data.pop('datos:j_id37')
            # validate also destiny postal_code
            fill_data['datos:j_id76'] = 'datos:j_id76'
            fill_data['datos:j_id75'] = destination.postal_code
            resp = self._client.post(self._urls['capture'], fill_data)
            if 'Código Postal válido' in resp.text:
                fill_data.pop('datos:j_id76')
                return fill_data
            else:
                raise DhlmexException('Invalid destiny postal code')
        else:
            raise DhlmexException('Invalid origin postal code')



from dataclasses import dataclass


@dataclass
class Origin:
    company: str
    contact: str
    email: str
    phone: str
    address1: str
    postal_code: str
    neighborhood: str
    city: str
    state: str



import re
from time import sleep
from typing import Dict, Optional, Tuple

from bs4 import BeautifulSoup
from requests import HTTPError, Response

from dhlmex.exceptions import DhlmexException
from dhlmex.resources.origin import Origin

from .base import Resource
from .destination import Destination
from .order_details import OrderDetails


class Guide(Resource):
    @classmethod
    def create_guide(
        cls, origin: Origin, destination: Destination, details: OrderDetails,
    ) -> Tuple[str, Optional[bytes]]:
        guide = cls()
        try:
            guides_data = guide._get_guide_data()
            if guides_data:
                guide._select_guide(guides_data)
                view_state = guide._fill_guide_table(
                    origin, destination, details
                )
                guide._confirm_capture(view_state)
                guide_number = guide._force_percent(view_state)
                guide_bytes = guide._download_pdf(guide_number)
                return guide_number, guide_bytes
            else:
                raise DhlmexException('No available guides')
        except HTTPError as httpe:
            raise httpe

    def _get_guide_data(self) -> Dict:
        resp = self._client.post(self._urls['home'], {})
        data = self.get_data(resp, self._actions['print'])
        soup = BeautifulSoup(resp.text, features='html.parser')
        field = soup.find('input', id=re.compile('panelBarInput')).attrs[
            'name'
        ]
        data[field] = self._actions['print']['code']
        guides_data = {}
        resp = self._client.post(self._urls['home'], data)
        if 'seleccionar' in resp.text:
            soup = BeautifulSoup(resp.text, features='html.parser')
            view_state = soup.find('input', id='javax.faces.ViewState').attrs[
                'value'
            ]
            table = soup.find('table', id='j_id6:pnlOrdenesEncontradas')
            table_body = table.find('tbody')
            js = table_body.find('a', text='seleccionar').attrs['onclick']
            matches = re.findall(r"\'(.+?)\'", js)
            form_ids = [match for match in matches if match.startswith('j_id')]
            j_pair_id = form_ids[1]
            j_id = form_ids[0]
            guides_data = {
                'AJAXREQUEST': '_viewRoot',
                j_id: j_id,
                'javax.faces.ViewState': view_state,
                j_pair_id: j_pair_id,
            }
        return guides_data

    def _select_guide(self, guides_data: Dict) -> Dict:
        resp = self._client.post(self._urls['guide'], guides_data)
        soup = BeautifulSoup(resp.text, features='html.parser')
        view_state = soup.find('input', id='javax.faces.ViewState').attrs[
            'value'
        ]
        j_id2 = soup.find('input', value='').attrs['name']
        js = soup.find('input', value='').attrs['onkeyup']
        matches = re.findall(r"\'(.+?)\'", js)
        form_ids = [match for match in matches if match.startswith('j_id')]
        j_pair_id = form_ids[1]
        j_id = form_ids[0]
        select_data = {
            'AJAXREQUEST': '_viewRoot',
            j_id: j_id,
            j_id2: '1',
            'javax.faces.ViewState': view_state,
            j_pair_id: j_pair_id,
            'AJAX:EVENTS_COUNT': '1',
        }
        self._client.post(self._urls['guide'], select_data)
        select_data.pop(j_pair_id, None)
        select_data.pop('AJAX:EVENTS_COUNT', None)
        select_data[
            'j_id6:btnGuardarCotizacion'
        ] = 'j_id6:btnGuardarCotizacion'
        self._client.post(self._urls['guide'], select_data)
        return select_data

    def _fill_guide_table(
        self, origin: Origin, destination: Destination, details: OrderDetails
    ) -> str:
        resp = self._client.get(self._urls['capture'])
        soup = BeautifulSoup(resp.text, features='html.parser')
        view_state = soup.find('input', id='javax.faces.ViewState').attrs[
            'value'
        ]
        fill_data = self._client.post_codes.validate_postal_codes(
            origin, destination, view_state
        )
        fill_data['datos:j_id15'] = origin.company
        fill_data['datos:j_id19'] = origin.contact
        fill_data['datos:emailOrigen'] = origin.email
        fill_data['datos:j_id24'] = origin.phone
        fill_data['datos:j_id28'] = origin.address1
        fill_data['datos:j_id36'] = origin.postal_code
        fill_data['datos:j_id41'] = origin.neighborhood
        fill_data['datos:j_id45'] = origin.city
        fill_data['datos:j_id49'] = origin.state
        fill_data['datos:j_id54'] = destination.company
        fill_data['datos:j_id58'] = destination.contact
        fill_data['datos:emailDestino'] = destination.email
        fill_data['datos:j_id63'] = destination.phone
        fill_data['datos:j_id67'] = destination.address1
        fill_data['datos:j_id75'] = destination.postal_code
        fill_data['datos:j_id80'] = destination.neighborhood
        fill_data['datos:j_id84'] = destination.city
        fill_data['datos:j_id88'] = destination.state
        fill_data['datos:j_id71'] = details.description
        fill_data['datos:j_id93'] = details.content
        fill_data['javax.faces.ViewState'] = view_state
        fill_data['datos:j_id105'] = 'datos:j_id105'

        msg = self._validate_data(fill_data)
        if msg:
            raise DhlmexException(f'Invalid data: {msg}')
        else:
            return fill_data['javax.faces.ViewState']

    def _validate_data(self, fill_data: Dict) -> str:
        resp = self._client.post(self._urls['capture'], fill_data)
        msg = ''
        if 'messageError' in resp.text:
            soup = BeautifulSoup(resp.text, features='html.parser')
            msg = soup.find('span', {'class': 'rich-messages-label'}).text
        return msg

    def _confirm_capture(self, view_state: str) -> Response:
        confirm_data = {
            'AJAXREQUEST': '_viewRoot',
            'j_id109': 'j_id109',
            'javax.faces.ViewState': view_state,
            'j_id109:j_id112': 'j_id109:j_id112',
        }
        return self._client.post(self._urls['capture'], confirm_data)

    def _force_percent(self, view_state: str, retries: int = 20) -> str:
        force_data = {
            'AJAXREQUEST': '_viewRoot',
            'j_id115': 'j_id115',
            'javax.faces.ViewState': view_state,
            'j_id115:pb_sub': 'j_id115:pb_sub',
            'forcePercent': 'complete',
            'ajaxSingle': 'j_id115:pb_sub',
        }
        guide_number = ''
        while retries:
            resp = self._client.post(self._urls['capture'], force_data)
            if 'Procesada correctamente' in resp.text:
                soup = BeautifulSoup(resp.text, features='html.parser')
                guide_number = soup.find(
                    'td', id='j_id115:tblElementos:0:j_id123'
                ).text
                break
            else:
                sleep(1)
                retries -= 1
        if retries == 0:
            raise DhlmexException('Error while capturing guide data')
        return guide_number

    def _move_page(self, view_state: str, page: str) -> Response:
        final_data = {
            'AJAXREQUEST': '_viewRoot',
            'j_id6': 'j_id6',
            'javax.faces.ViewState': view_state,
            'ajaxSingle': 'j_id6:j_id37',
            'j_id6:j_id37': page,
            'AJAX:EVENTS_COUNT': '1',
        }
        return self._client.post(self._urls['print'], final_data)

    def _download_pdf(self, guide_number: str) -> Optional[bytes]:
        resp = self._client.post(self._urls['home'], {})
        data = self.get_data(resp, self._actions['download'])
        resp = self._client.post(self._urls['home'], data)
        if guide_number not in resp.text:  # search on last page
            soup = BeautifulSoup(resp.text, features='html.parser')
            view_state = soup.find('input', id='javax.faces.ViewState').attrs[
                'value'
            ]
            pages = len(
                soup.find("div", {"class": "rich-datascr"}).find_all('td')
            )
            resp = self._move_page(view_state, 'last')
            for _ in range(pages):
                if guide_number not in resp.text:  # search previous pages
                    resp = self._move_page(view_state, 'previous')
                else:
                    break
        if guide_number not in resp.text:
            raise DhlmexException(f'Guide {guide_number} not found')
        soup = BeautifulSoup(resp.text, features='html.parser')
        view_state = soup.find('input', id='javax.faces.ViewState').attrs[
            'value'
        ]
        td = soup.find('td', text=guide_number)
        tds = [td for td in td.next_siblings]
        j_pair_id = tds[-1].find('a').attrs['id']

        guide_data = {
            'AJAXREQUEST': '_viewRoot',
            'j_id6': 'j_id6',
            'javax.faces.ViewState': view_state,
            j_pair_id: j_pair_id,
        }
        self._client.post(self._urls['print'], guide_data)
        resp = self._client.get(self._urls['pdf'])
        if resp.ok:
            return resp.content
        else:
            return None



import re
from typing import ClassVar, Dict

from bs4 import BeautifulSoup
from requests import Response


class Resource:
    _client: ClassVar["dhlmex.Client"]  # type: ignore
    _urls: Dict[str, str] = {
        'login': '/jsp/app/login/login.xhtml',
        'home': '/jsp/app/inicio/inicio.xhtml',
        'guide': '/jsp/app/cliente/impresionClienteSubUsuario.xhtml',
        'capture': '/jsp/app/cliente/capturaDatosImpresionClienteSU.xhtml',
        'print': '/jsp/app/cliente/guiasImpresas.xhtml',
        'pdf': '/generaImpresionPDF',
    }
    _actions: Dict[str, Dict[str, str]] = {
        'close': {
            'text': 'Cerrar Sesión',
            'code': 'j_id9:j_id26',
            'end': 'j_id9:j_id30',
        },
        'print': {
            'text': 'Impresión Sub Usuario',
            'code': 'j_id9:j_id14',
            'end': 'j_id9:j_id16',
        },
        'download': {
            'text': 'Guías Impresas',
            'code': 'j_id9:j_id18',
            'end': 'j_id9:j_id10',
        },
    }

    @staticmethod
    def get_data(resp: Response, action: Dict) -> Dict:
        soup = BeautifulSoup(resp.text, features='html.parser')
        view_state = soup.find('input', id='javax.faces.ViewState').attrs[
            'value'
        ]
        js = soup.find('a', text=action['text']).attrs['onclick']
        matches = re.findall(r"\'(.+?)\'", js)
        form_ids = [match for match in matches if match.startswith('j_id')]
        j_pair_id = form_ids[1].split(',')[0]
        j_id = form_ids[0]

        return {
            j_id: j_id,
            j_pair_id: action['code'],
            'javax.faces.ViewState': view_state,
            action['end']: action['end'],
        }
